@@ -12,7 +12,7 @@ WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
-this program.  If not, see <http://www.gnu.org/licenses/>.  */
+this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #include "makeint.h"
 #include "filedef.h"
@@ -54,6 +54,52 @@ make_toui (const char *str, const char **error)
   return val;
 }
 
+/* Convert val into a string, written to buf.  buf must be large enough
+   to hold the largest possible value, plus a nul byte.  Returns buf.
+   We can't use standard PRI* here: those are based on intNN_t types.  */
+
+char *
+make_lltoa (long long val, char *buf)
+{
+  sprintf (buf, "%" MK_PRI64_PREFIX "d", val);
+  return buf;
+}
+
+char *
+make_ulltoa (unsigned long long val, char *buf)
+{
+  sprintf (buf, "%" MK_PRI64_PREFIX "u", val);
+  return buf;
+}
+
+/* Simple random number generator, for use with shuffle.
+   This doesn't need to be truly random, just pretty random.  Use our own
+   implementation rather than relying on the C runtime's rand() so we always
+   get the same results for a given seed, regardless of C runtime.  */
+
+static unsigned int mk_state = 0;
+
+void
+make_seed (unsigned int seed)
+{
+  mk_state = seed;
+}
+
+unsigned int
+make_rand ()
+{
+  /* mk_state must never be 0.  */
+  if (mk_state == 0)
+    mk_state = (unsigned int)(time (NULL) ^ make_pid ()) + 1;
+
+  /* A simple xorshift RNG.  */
+  mk_state ^= mk_state << 13;
+  mk_state ^= mk_state >> 17;
+  mk_state ^= mk_state << 5;
+
+  return mk_state;
+}
+
 /* Compare strings *S1 and *S2.
    Return negative if the first is less, positive if it is greater,
    zero if they are equal.  */
@@ -484,6 +530,22 @@ spin (const char* type)
     }
 }
 
+void
+dbg (const char *fmt, ...)
+{
+  FILE *fp = fopen ("/tmp/gmkdebug.log", "a+");
+  va_list args;
+  char buf[4096];
+
+  va_start (args, fmt);
+  vsprintf (buf, fmt, args);
+  va_end (args);
+
+  fprintf(fp, "%u: %s\n", (unsigned) make_pid (), buf);
+  fflush (fp);
+  fclose (fp);
+}
+
 #endif
 
 
@@ -506,33 +568,49 @@ umask (mode_t mask)
 }
 #endif
 
-#define MAKE_TMPDIR         "MAKE_TMPDIR"
 #ifdef VMS
 # define DEFAULT_TMPFILE    "sys$scratch:gnv$make_cmdXXXXXX.com"
-# define DEFAULT_TMPDIR     "/sys$scratch/"
 #else
-# define DEFAULT_TMPFILE     "GmXXXXXX"
-# ifdef P_tmpdir
-#  define DEFAULT_TMPDIR    P_tmpdir
-# else
-#  define DEFAULT_TMPDIR    "/tmp"
-# endif
+# define DEFAULT_TMPFILE    "GmXXXXXX"
 #endif
 
-static const char *
+const char *
 get_tmpdir ()
 {
   static const char *tmpdir = NULL;
 
   if (!tmpdir)
     {
-      if (((tmpdir = getenv (MAKE_TMPDIR)) == NULL || *tmpdir == '\0')
 #if defined (__MSDOS__) || defined (WINDOWS32) || defined (__EMX__)
-          && ((tmpdir = getenv ("TMP")) == NULL || *tmpdir == '\0')
-          && ((tmpdir = getenv ("TEMP")) == NULL || *tmpdir == '\0')
+# define TMP_EXTRAS   "TMP", "TEMP",
+#else
+# define TMP_EXTRAS
 #endif
-          && ((tmpdir = getenv ("TMPDIR")) == NULL || *tmpdir == '\0'))
-        tmpdir = DEFAULT_TMPDIR;
+      const char *tlist[] = { "MAKE_TMPDIR", "TMPDIR", TMP_EXTRAS NULL };
+      const char **tp;
+      unsigned int found = 0;
+
+      for (tp = tlist; *tp; ++tp)
+        if ((tmpdir = getenv (*tp)) && *tmpdir != '\0')
+          {
+            struct stat st;
+            int r;
+            found = 1;
+            EINTRLOOP(r, stat (tmpdir, &st));
+            if (r < 0)
+              OSSS (error, NILF,
+                    _("%s value %s: %s"), *tp, tmpdir, strerror (errno));
+            else if (! S_ISDIR (st.st_mode))
+              OSS (error, NILF,
+                   _("%s value %s: not a directory"), *tp, tmpdir);
+            else
+              return tmpdir;
+          }
+
+      tmpdir = DEFAULT_TMPDIR;
+
+      if (found)
+        OS (error, NILF, _("using default temporary directory '%s'"), tmpdir);
     }
 
   return tmpdir;
@@ -543,44 +621,45 @@ get_tmptemplate ()
 {
   const char *tmpdir = get_tmpdir ();
   char *template;
-  size_t len;
+  char *cp;
 
-  len = strlen (tmpdir);
-  template = xmalloc (len + CSTRLEN (DEFAULT_TMPFILE) + 2);
-  strcpy (template, tmpdir);
+  template = xmalloc (strlen (tmpdir) + CSTRLEN (DEFAULT_TMPFILE) + 2);
+  cp = stpcpy (template, tmpdir);
 
-#ifdef HAVE_DOS_PATHS
-  if (template[len - 1] != '/' && template[len - 1] != '\\')
-    strcat (template, "/");
-#else
-# ifndef VMS
-  if (template[len - 1] != '/')
-    strcat (template, "/");
-# endif /* !VMS */
-#endif /* !HAVE_DOS_PATHS */
+#if !defined VMS
+  /* It's not possible for tmpdir to be empty.  */
+  if (! ISDIRSEP (cp[-1]))
+    *(cp++) = '/';
+#endif
 
-  strcat (template, DEFAULT_TMPFILE);
+  strcpy (cp, DEFAULT_TMPFILE);
 
   return template;
 }
 
-char *
+#if !HAVE_MKSTEMP || !HAVE_FDOPEN
+/* Generate a temporary filename.  This is not safe as another program could
+   snipe our filename after we've generated it: use this only on systems
+   without more secure alternatives.  */
+
+static char *
 get_tmppath ()
 {
   char *path;
 
-#ifdef HAVE_MKTEMP
-  path = get_tmptemplate();
+# ifdef HAVE_MKTEMP
+  path = get_tmptemplate ();
   if (*mktemp (path) == '\0')
     pfatal_with_name ("mktemp");
-#else
+# else
   path = xmalloc (L_tmpnam + 1);
   if (tmpnam (path) == NULL)
     pfatal_with_name ("tmpnam");
-#endif
+# endif
 
   return path;
 }
+#endif
 
 /* Generate a temporary file and return an fd for it.  If name is NULL then
    the temp file is anonymous and will be deleted when the process exits.  */
@@ -589,7 +668,6 @@ get_tmpfd (char **name)
 {
   int fd = -1;
   char *tmpnm;
-  mode_t mask;
 
   /* If there's an os-specific way to get an anoymous temp file use it.  */
   if (!name)
@@ -599,9 +677,6 @@ get_tmpfd (char **name)
         return fd;
     }
 
-  /* Preserve the current umask, and set a restrictive one for temp files.  */
-  mask = umask (0077);
-
 #if defined(HAVE_MKSTEMP)
   tmpnm = get_tmptemplate ();
 
@@ -613,48 +688,61 @@ get_tmpfd (char **name)
   /* Can't use mkstemp(), but try to guard against a race condition.  */
   EINTRLOOP (fd, open (tmpnm, O_CREAT|O_EXCL|O_RDWR, 0600));
 #endif
-
-  umask (mask);
+  if (fd < 0)
+    OSS (fatal, NILF,
+         _("create temporary file %s: %s"), tmpnm, strerror (errno));
 
   if (name)
     *name = tmpnm;
   else
     {
-      unlink (tmpnm);
+      int r;
+      EINTRLOOP (r, unlink (tmpnm));
+      if (r < 0)
+        OSS (fatal, NILF,
+             _("unlink temporary file %s: %s"), tmpnm, strerror (errno));
       free (tmpnm);
     }
 
   return fd;
 }
 
+/* Return a FILE* for a temporary file, opened in the safest way possible.
+   Set name to point to an allocated buffer containing the name of the file.
+   Note, this cannot be NULL!  */
 FILE *
 get_tmpfile (char **name)
 {
+  /* Be consistent with tmpfile, which opens as if by "wb+".  */
+  const char *tmpfile_mode = "wb+";
+  FILE *file;
+
 #if defined(HAVE_FDOPEN)
   int fd = get_tmpfd (name);
 
-  return fd < 0 ? NULL : fdopen (fd, "w");
+  ENULLLOOP (file, fdopen (fd, tmpfile_mode));
+  if (file == NULL)
+    OSS (fatal, NILF,
+         _("fdopen: temporary file %s: %s"), *name, strerror (errno));
 #else
   /* Preserve the current umask, and set a restrictive one for temp files.  */
   mode_t mask = umask (0077);
+  int err;
 
-  char *tmpnm = get_tmppath ();
+  *name = get_tmppath ();
 
-  /* Not secure, but...?  If name is NULL we could use tmpfile()...  */
-  FILE *file = fopen (tmpnm, "w");
+  /* Although this fopen is insecure, it is executed only on non-fdopen
+     platforms, which should be a rarity nowadays.  */
+
+  ENULLLOOP (file, fopen (*name, tmpfile_mode));
+  if (file == NULL)
+    OSS (fatal, NILF,
+         _("fopen: temporary file %s: %s"), *name, strerror (errno));
 
   umask (mask);
-
-  if (name)
-    *name = tmpnm;
-  else
-    {
-      unlink (tmpnm);
-      free (tmpnm);
-    }
+#endif
 
   return file;
-#endif
 }
 
 
@@ -688,7 +776,7 @@ strcasecmp (const char *s1, const char *s2)
    substitute for it, define our own version.  */
 
 int
-strncasecmp (const char *s1, const char *s2, int n)
+strncasecmp (const char *s1, const char *s2, size_t n)
 {
   while (n-- > 0)
     {
@@ -719,11 +807,11 @@ get_path_max (void)
 
   if (value == 0)
     {
-      long int x = pathconf ("/", _PC_PATH_MAX);
+      long x = pathconf ("/", _PC_PATH_MAX);
       if (x > 0)
-        value = x;
+        value = (unsigned int) x;
       else
-        return MAXPATHLEN;
+        value = PATH_MAX;
     }
 
   return value;
